/**
 * Title:        R-S Decoder Applet<p>
 * Project:      EE/Ma 127b Error-Correcting Codes
 *               Project 1 - Reed-Solomon Decoder<p>
 * Copyright:    Copyright (c) Ling Li, 2001<p>
 * Company:      California Institute of Technology<p>
 * @author Ling Li
 * @email ling@cs.caltech.edu
 * @version 1.0
 */

import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import ecc.*;

public class RSDecoder extends Applet {
    GF gf;
    RS rs;
    static int p = 2, n = 31, r = 16;
    static int genpoly[] = {1, 0, 1, 0, 0, 1};
    static char eraChar = '*';

    GridBagLayout gridbag = new GridBagLayout();
    Label lblReceived = new Label();
    Label lblDecoded = new Label();
    TextField txtReceived = new TextField();
    TextField txtDecoded = new TextField();
    TextArea txtComment = new TextArea();
    Button btnClearI = new Button();
    Button btnClearO = new Button();
    Button btnDecode = new Button();

    //Construct the applet
    public RSDecoder() {
        gf = new GF(p, genpoly);
        rs = new RS(gf, r);
    }

    //Initialize the applet
    public void init() {
        try {
            jbInit();
            txtReceived.setText("");
            txtDecoded.setText("");
            //txtComment.setText("");
        }
        catch(Exception e) {
            e.printStackTrace();
        }
    }

    //Component initialization
    private void jbInit() throws Exception {
        this.setLayout(gridbag);
        lblReceived.setText("Received code R:");
        lblDecoded.setText("Decoded codeword C:");
        txtDecoded.setEditable(false);
        txtComment.setEditable(false);
        btnClearI.setLabel("Reset");
        btnClearO.setLabel("Clear");
        btnClearI.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                btnClearI_mouseClicked(e);
            }
        });
        btnClearO.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                btnClearO_mouseClicked(e);
            }
        });
        btnDecode.setLabel("Decode");
        btnDecode.addMouseListener(new java.awt.event.MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                btnDecode_mouseClicked(e);
            }
        });

        GridBagConstraints c = new GridBagConstraints();

        c.gridx = c.gridy = 3;
        c.fill = c.HORIZONTAL;
        c.insets = new Insets(0, 10, 0, 10);
        gridbag.setConstraints(btnClearO, c);
        add(btnClearO);

        c.gridy = 2;
        gridbag.setConstraints(btnClearI, c);
        add(btnClearI);

        c.gridy = 1;
        gridbag.setConstraints(btnDecode, c);
        add(btnDecode);

        c.gridx = c.gridy = 0; c.ipadx = 5;
        c.anchor = c.NORTHWEST; c.fill = c.NONE;
        gridbag.setConstraints(lblReceived, c);
        add(lblReceived);

        c.gridy = 2;
        gridbag.setConstraints(lblDecoded, c);
        add(lblDecoded);

        c.gridx = 0; c.gridy = 3;
        c.gridwidth = 3; c.weightx = 1.0;
        c.fill = c.HORIZONTAL;
        gridbag.setConstraints(txtDecoded, c);
        add(txtDecoded);

        c.gridy = 1;
        gridbag.setConstraints(txtReceived, c);
        add(txtReceived);

        c.gridx = 0; c.gridy = 4; c.gridwidth = 4;
        c.gridheight = 2; c.weighty = 0.5;
        c.fill = c.BOTH; c.anchor = c.CENTER;
        c.insets = new Insets(10, 5, 5, 5);
        c.ipadx = c.ipady = 5;
        gridbag.setConstraints(txtComment, c);
        add(txtComment);
    }

    //Get Applet information
    public String getAppletInfo() {
        return "R-S Decoder. Ling Li, Caltech, 2001.";
    }

    //Get parameter info
    public String[][] getParameterInfo() {
        return null;
    }

    private int parseString(String s, int c[]) {
        s = s.trim();
        int len = s.length();
        int clen = c.length;
        int i = 0;

        if (s.indexOf(',') > 0) {
            int pos = 0;
            // example: 32, 23, 1, ...
            try {
                while (i < clen && pos < len) {
                    int po2 = s.indexOf(',', pos);
                    if (po2 < 0) po2 = len;
                    String ssub = s.substring(pos, po2).trim();
                    if (ssub.charAt(0) == eraChar) c[i] = -1;
                    else c[i] = Integer.parseInt(ssub);
                    if (c[i] < -1 || c[i] > n) break;

                    i++; pos = po2+1;
                }
            } catch (NumberFormatException e) {
                //System.out.print("Format error\n");
            }
        }
        else {
            byte b[] = s.toUpperCase().getBytes();
            // example "ABCD2"
            for (; i < clen && i < len; i++)
                if (s.charAt(i) == eraChar) c[i] = -1;
                else {
                    c[i] = b[i] - 65;
                    if (c[i] < 0) {
                        c[i] = c[i] + 42;
                        if (c[i] < 26 || c[i] > n) break;
                    } else if (c[i] > 25) break;
                }
        }
        return i;
    }

    void btnClearI_mouseClicked(MouseEvent e) {
        txtReceived.setText("");
        txtDecoded.setText("");
    }

    void btnClearO_mouseClicked(MouseEvent e) {
        txtDecoded.setText("");
        txtComment.setText("");
    }

    void btnDecode_mouseClicked(MouseEvent e) {
        String recCode = txtReceived.getText();

        int rec[] = new int[n];
        int k = parseString(recCode, rec);
        if (k < n) {
            txtComment.setText(txtComment.getText() + "Something wrong: only "
                  + Integer.toString(k) + " symbol(s) correctly received\n");
            txtDecoded.setText("Error");
        }
        else {
            String s = txtComment.getText();
            s += "You entered:\n  ";
            if (rec[0] < 0) s += "*";
            else s += String.valueOf(rec[0]);
            for (int i = 1; i < k; i++)
                if (rec[i] < 0) s += ",*";
                else s += "," + String.valueOf(rec[i]);

            int ret = rs.decode_time(rec);
            if (ret >= 0)
                txtDecoded.setText(rs.C.toMatrixString(n));
            else txtDecoded.setText("Can't decode");
            while (true) {
                if (ret == -5) { s += "\n  Too many erasures. quit"; break; }
                if (ret ==  0) { s += "\n  Syndrome is 0"; break; }
                                 s += "\n  Syndrome is " + rs.S.toString();
                                 s += "\nCalculating sigma & omega:";
                if (ret == -1) { s += "\n  Can't get sigma: v0 == 0"; break; }
                                 s += "\n  sigma = " + rs.sigma.toString()
                                    + "\n  omega = " + rs.omega.toString();
                                 s += "\n    gcd(s,o) = " +
                                   GFPoly.euclid(rs.sigma, rs.omega).toString();
                                 s += ", e0 = " + Integer.toString(rs.e0) +
                                      ", e1 = " + Integer.toString(rs.e1) +
                                      " (guess)";
                if (ret == -2) { s += "\n  deg(sigma) <= deg(omega)."; break; }
                                 s += "\nDecoding in time domain:";
                                 s += "\n  sigma'= " + rs.sigma_d.toString();
                if (ret == -3)   s += "\n  Too few roots of sigma";
                else             s += "\n  E = " + rs.E.toMatrixString(n);

                                 s += "\nDecoding in frequency domain:";
                ret = rs.decode_freq(rec);
                if (ret == -4)   s += "\n  S verified error";
                else if (ret > 0)s += "\n  E = " + rs.E.toMatrixString(n);
                break;
            }

            txtComment.setText(s + "\n\n");
        }
    }
}