/**
 * Title:        Polynomials over Galois Field<p>
 * Package:      ecc (Error-Correcting Codes)
 * Project:      EE/Ma 127b Error-Correcting Codes
 *               Project 1 - Reed-Solomon Decoder<p>
 * Copyright:    Copyright (c) Ling Li, 2001<p>
 * Company:      California Institute of Technology<p>
 * @author Ling Li
 * @email ling@cs.caltech.edu
 * @version 1.1
 */

/** TODO:
 *    1. clean the code
 *    2. GFPoly as a sub-class of GF?
 *    3. How to share gf between different instances?
 *    4. functions...
 */

package ecc;

public class GFPoly {
    private GF gf;
    private int c[];    // coefficients
    private int deg;    // degree of the polynomial
    protected GFPoly u1, v1, r1;

    public GFPoly() {
        deg = 0;
        c = new int[1];
        gf = new GF();
    }

    public GFPoly(GF cgf, int c0) {
        gf = (GF) cgf.clone();
        setC(c0);
    }

    public GFPoly(GF cgf, int coeff[]) {
        gf = (GF) cgf.clone();
        setC(coeff);
    }

    public Object clone() {
        GFPoly t = new GFPoly(gf, c);
        return (Object) t;
    }

    public void setC(int coeff[]) {
        c = (int[]) coeff.clone();
        for (deg = c.length-1; deg > 0; deg--)
            if (c[deg] != 0) break;
    }

    public void setC(GFPoly gp) {
        setC(gp.c);
    }

    public void setC(int c0) {
        c = new int[1];
        c[0] = c0; deg = 0;
    }

    public int[] getC() {
        return (int[]) c.clone();
    }

    public int getC(int n) {
        if (n < 0 || n > deg) return 0;
        else return c[n];
    }

    public void setDegree(int d) {
        //assert(d >= 0);
        if (d < deg) {
            for (deg = d; deg > 0; deg--)
                if (c[deg] != 0) break;
        }
    }

    public int getDegree() {
        return deg;
    }

    public GFPoly add(GFPoly b) {
        int d = b.deg, i;
        if (d < deg) d = deg;
        int sum[] = new int[d+1];

        for (i = 0; i <= deg; i++)
            sum[i] = gf.add(sum[i], c[i]);
        for (i = 0; i <= b.deg; i++)
            sum[i] = gf.add(sum[i], b.c[i]);

        return (new GFPoly(gf, sum));
    }

    public GFPoly sub(GFPoly b) {
        int d = b.deg, i;
        if (d < deg) d = deg;
        int sum[] = new int[d+1];

        for (i = 0; i <= deg; i++)
            sum[i] = gf.sub(sum[i], c[i]);
        for (i = 0; i <= b.deg; i++)
            sum[i] = gf.sub(sum[i], b.c[i]);

        return (new GFPoly(gf, sum));
    }

    public GFPoly mul(GFPoly b) {
        int d = deg + b.deg;
        int prod[] = new int[d+1];

        for (int i = 0; i <= deg; i++)
            for (int j = 0; j <= b.deg; j++)
                prod[i+j] = gf.add(prod[i+j], gf.mul(c[i], b.c[j]));

        return (new GFPoly(gf, prod));
    }

    public GFPoly mul(int b) {
        if (b == 0)
            return (new GFPoly(gf, 0));

        int prod[] = new int[deg+1];

        for (int i = 0; i <= deg; i++)
            prod[i] = gf.mul(c[i], b);

        return (new GFPoly(gf, prod));
    }

    public GFPoly div(GFPoly b) {
        if (deg < b.deg)
            return (new GFPoly(gf, 0));

        int d = deg - b.deg;
        int invb = gf.inv(b.c[b.deg]);
        int rem[] = (int[]) c.clone();
        int quo[] = new int[d+1];

        for (int i = d; i >= 0; i--) {
            quo[i] = gf.mul(rem[i+b.deg], invb);
            for (int j = 0; j <= b.deg; j++)
                rem[j+i] = gf.sub(rem[j+i], gf.mul(quo[i], b.c[j]));
        }

        return (new GFPoly(gf, quo));
    }

    public GFPoly div(int b) {
        return (this.mul(gf.inv(b)));
    }

    public String toString() {
        if (deg == 0 && c[0] == 0)
            return "0";

        String s = "";
        for (int i = 0; i <= deg; i++)
            if (c[i] != 0) {
                if (c[i] != 1 || i == 0)
                    s += Integer.toString(c[i]);
                if (i == 1)
                    s += "x";
                else if (i > 1)
                    s += "x^" + Integer.toString(i);
                if (i < deg)
                    s += " + ";
            }
        return s;
    }

    public String toMatrixString(int len) {
        int i;
        String s = String.valueOf(c[0]);
        for (i = 1; i < c.length; i++)
            s += "," + String.valueOf(c[i]);
        for (i = c.length; i < len; i++)
            s += ",0";
        return s;
    }

    // evaluate f(alpha^pow) where f(x) = \sum c[i]*x^i
    public int evalfun(int x) {
        int f = 0, xi = gf.power(0);
        for (int i = 0; i <= deg; i++, xi = gf.mul(xi, x))
            f = gf.add(f, gf.mul(c[i], xi));
        return f;
    }

    public GFPoly fft() {
        return fft(c.length);
    }

    public GFPoly fft(int k) {
        int freq[] = new int[k];
        for (int i = 0; i < k; i++)
            freq[i] = evalfun(gf.power(i));
        return (new GFPoly(gf, freq));
    }

    public GFPoly ifft() {
        int time[] = new int[c.length];
        int n = gf.nOne(time.length);
        if (n != 0)
            for (int i = 0; i < time.length; i++)
                time[i] = gf.div(evalfun(gf.power(-i)), n);
        return (new GFPoly(gf, time));
    }

    public void euclid(GFPoly a, GFPoly b, int degv, int degr) {
        GFPoly u0, v0, r0, q, temp;

        u0 = new GFPoly(gf, 1); u1 = new GFPoly(gf, 0);
        v0 = new GFPoly(gf, 0); v1 = new GFPoly(gf, 1);
        r0 = (GFPoly) a.clone(); r1 = (GFPoly) b.clone();

        while (r1.deg > degr || v1.deg > degv) {
            q = r0.div(r1);

            temp = u0.sub(q.mul(u1));
            u0 = u1; u1 = temp;

            temp = v0.sub(q.mul(v1));
            v0 = v1; v1 = temp;

            temp = r0.sub(q.mul(r1));
            r0 = r1; r1 = temp;
        }
    }

    public static GFPoly euclid(GFPoly a, GFPoly b) {
        GFPoly u0, v0, r0, u1, v1, r1, q, temp;

        u0 = new GFPoly(a.gf, 1); u1 = new GFPoly(a.gf, 0);
        v0 = new GFPoly(a.gf, 0); v1 = new GFPoly(a.gf, 1);
        r0 = (GFPoly) a.clone(); r1 = (GFPoly) b.clone();

        while (r1.deg != 0 || r1.c[0] != 0) {
            q = r0.div(r1);

            temp = u0.sub(q.mul(u1));
            u0 = u1; u1 = temp;

            temp = v0.sub(q.mul(v1));
            v0 = v1; v1 = temp;

            temp = r0.sub(q.mul(r1));
            r0 = r1; r1 = temp;
        }
        return r0;
    }

    public GFPoly deriv() {
        int de[] = new int[deg];
        for (int i = 0; i < deg; i++)
            de[i] = gf.mul(c[i+1], gf.nOne(i+1));
        return (new GFPoly(gf, de));
    }
}