/**
 * Title:        Reed-Solomon Decoder<p>
 * Package:      ecc (Error-Correcting Codes)
 * Project:      EE/Ma 127b Error-Correcting Codes
 *               Project 1 - Reed-Solomon Decoder<p>
 * Copyright:    Copyright (c) Ling Li, 2001<p>
 * Company:      California Institute of Technology<p>
 * @author Ling Li
 * @email ling@cs.caltech.edu
 * @version 1.0
 */

/** TODO:
 *    1. clean the code / more efficiency?
 *    2. Encoder?
 *    3. avoid variables exposure, such as E/C/R/S/sigma/omega...
 */

package ecc;

public class RS {
    protected GF gf;
    protected int n, r;
    public GFPoly R, S, E, C, sigma, omega, sigma_d;
    public int e0, e1;

    public RS(GF _gf, int _r) {
        gf = (GF) _gf.clone();
        init(_r);
    }

    public RS(int p, int genpoly[], int _r) {
        gf = new GF(p, genpoly);
        init(_r);
    }

    private void init(int _r) {
        r = _r;
        n = gf.max_num();
        R = new GFPoly(gf, 0);
        E = new GFPoly(gf, 0);
        C = new GFPoly(gf, 0);
    }

    public int getSigma(int received[]) {
        int i;
        int rec[] = (int[]) received.clone();
        GFPoly sigma0 = new GFPoly(gf, 1);
        for (i = 0, e0 = 0; i < rec.length; i++) {
            if (rec[i] < 0) {
                rec[i] = 0;
                int erl[] = {1, gf.sub(0, gf.power(i))};
                sigma0 = sigma0.mul(new GFPoly(gf, erl));
                e0++;
            }
        }
        if (e0 > r) return -5;  // too many erasures
        R.setC(rec);

        int x[] = {0, 1};
        S = R.fft(r+1).div(new GFPoly(gf, x));

        if (S.getDegree() == 0 && S.getC(0) == 0) {
            E.setC(0); C.setC(R);
            return 0;
        }

        GFPoly S0 = sigma0.mul(S);
        S0.setDegree(r-1);  // mod x^r, maybe not necessary

        int xrc[] = new int[r+1];
        xrc[r] = 1;
        S.euclid(new GFPoly(gf, xrc), S0, (r-e0)/2, (r+e0-1)/2);

        int v0 = S.v1.getC(0);
        if (v0 == 0) return -1;

        e1 = S.v1.getDegree();
        sigma = S.v1.div(v0).mul(sigma0);
        omega = S.r1.div(v0);
        if (sigma.getDegree() <= omega.getDegree())
            return -2;

        return 1;
    }

    public int decode_time(int rec[]) {
        int t = getSigma(rec);
        if (t <= 0) return t;

        sigma_d = sigma.deriv();

        int e[] = new int[n];
        int e01 = 0;
        for (int i = 0; i < n; i++) {
            int alp = gf.power(-i);
            if (sigma.evalfun(alp) == 0) {
                int sde = sigma_d.evalfun(alp);
                if (sde != 0)
                    e[i] = gf.sub(0, gf.div(omega.evalfun(alp), sde));
                else return -3;
                e01++;
            }
            else e[i] = 0;
        }
        if (e01 < sigma.getDegree()) return -3;
        E.setC(e);
        C = R.sub(E);

        return 1;
    }

    public int decode_freq(int rec[]) {
        int t = getSigma(rec), i;
        if (t <= 0) return t;

        int vhat[] = new int[n];
        for (i = 0; i< r; i++)
            vhat[i+1] = S.getC(i);

        int sigc[] = sigma.getC();
        int deg = sigma.getDegree();
        for (i = r+1; i <= n+r; i++) {
            int sum = 0;
            for (int j = 1; j <= deg; j++)
                sum = gf.sub(sum, gf.mul(sigc[j], vhat[(n+i-j)%n]));

            vhat[i % n] = sum;
        }

        for (i = 0; i < r; i++)
            if (vhat[i+1] != S.getC(i)) return -4;

        E = (new GFPoly(gf, vhat)).ifft();
        C = R.sub(E);

        return 1;
    }
}